@@ -1,7 +1,7 @@
 /* -*- c-basic-offset: 8; -*- */
 /* shout.c: Implementation of public libshout interface shout.h
  *
- *  Copyright (C) 2002-2003 the Icecast team <team@icecast.org>
+ *  Copyright (C) 2002-2004 the Icecast team <team@icecast.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -16,6 +16,8 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ * $Id: shout.c 8808 2005-01-29 19:40:26Z brendan $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -36,9 +38,23 @@
 #include "util.h"
 
 /* -- local prototypes -- */
-static int login_xaudiocast(shout_t *self);
-static int login_icy(shout_t *self);
-static int login_http_basic(shout_t *self);
+static int queue_data(shout_queue_t *queue, const unsigned char *data, size_t len);
+static int queue_str(shout_t *self, const char *str);
+static int queue_printf(shout_t *self, const char *fmt, ...);
+static inline void queue_free(shout_queue_t *queue);
+static int send_queue(shout_t *self);
+static int get_response(shout_t *self);
+static int try_connect (shout_t *self);
+static int try_write (shout_t *self, const void *data, size_t len);
+
+static int create_request(shout_t *self);
+static int create_http_request(shout_t *self);
+static int create_xaudiocast_request(shout_t *self);
+static int create_icy_request(shout_t *self);
+static int parse_response(shout_t *self);
+static int parse_http_response(shout_t *self);
+static int parse_xaudiocast_response(shout_t *self);
+
 static char *http_basic_authorization(shout_t *self);
 
 /* -- static data -- */
@@ -127,61 +143,14 @@ int shout_open(shout_t *self)
 	/* sanity check */
 	if (!self)
 		return SHOUTERR_INSANE;
-
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return SHOUTERR_CONNECTED;
-
 	if (!self->host || !self->password || !self->port)
 		return self->error = SHOUTERR_INSANE;
-
-	if (self->format == SHOUT_FORMAT_VORBIS && self->protocol != SHOUT_PROTOCOL_HTTP)
+	if (self->format == SHOUT_FORMAT_OGG && self->protocol != SHOUT_PROTOCOL_HTTP)
 		return self->error = SHOUTERR_UNSUPPORTED;
 
-	if(self->protocol != SHOUT_PROTOCOL_HTTP) {
-		if (self->protocol == SHOUT_PROTOCOL_ICY)
-			self->socket = sock_connect(self->host, self->port+1);
-		else
-			self->socket = sock_connect(self->host, self->port);
-		if (self->socket <= 0)
-			return self->error = SHOUTERR_NOCONNECT;
-	}
-
-	if (self->protocol == SHOUT_PROTOCOL_HTTP) {
-		if ((self->error = login_http_basic(self)) != SHOUTERR_SUCCESS) {
-			return self->error;
-		}
-	} else if (self->protocol == SHOUT_PROTOCOL_XAUDIOCAST) {
-		if ((self->error = login_xaudiocast(self)) != SHOUTERR_SUCCESS) {
-			sock_close(self->socket);
-			return self->error;
-		}
-	} else if (self->protocol == SHOUT_PROTOCOL_ICY) {
-		if ((self->error = login_icy(self)) != SHOUTERR_SUCCESS) {
-			sock_close(self->socket);
-			return self->error;
-		}
-		
-	} else
-		return self->error = SHOUTERR_INSANE;
-  
-	if (self->format == SHOUT_FORMAT_VORBIS) {
-		if ((self->error = shout_open_vorbis(self)) != SHOUTERR_SUCCESS) {
-			sock_close(self->socket);
-			return self->error;
-		}
-	} else if (self->format == SHOUT_FORMAT_MP3) {
-		if ((self->error = shout_open_mp3(self)) != SHOUTERR_SUCCESS) {
-			sock_close(self->socket);
-			return self->error;
-		}
-	} else {
-		sock_close(self->socket);
-		return self->error = SHOUTERR_INSANE;
-	}
-
-	self->connected = 1;
-
-	return self->error;
+	return self->error = try_connect(self);
 }
 
 
@@ -190,14 +159,18 @@ int shout_close(shout_t *self)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (!self->connected)
+	if (self->state == SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_UNCONNECTED;
 
 	if (self->close)
 		self->close(self);
 
 	sock_close(self->socket);
-	self->connected = 0;
+	self->state = SHOUT_STATE_UNCONNECTED;
+	self->starttime = 0;
+	self->senttime = 0;
+	queue_free(&self->rqueue);
+	queue_free(&self->wqueue);
 
 	return self->error = SHOUTERR_SUCCESS;
 }
@@ -207,44 +180,63 @@ int shout_send(shout_t *self, const unsigned char *data, size_t len)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (!self->connected)
+	if (self->state != SHOUT_STATE_CONNECTED)
 		return self->error = SHOUTERR_UNCONNECTED;
 
 	if (self->starttime <= 0)
 		self->starttime = timing_get_time();
 
+	if (!len)
+		return send_queue(self);
+
 	return self->send(self, data, len);
 }
 
 ssize_t shout_send_raw(shout_t *self, const unsigned char *data, size_t len)
 {
 	ssize_t ret;
-	size_t remaining = len;
 
 	if (!self) 
 		return SHOUTERR_INSANE;
 
-	if (!self->connected)
+	if (self->state != SHOUT_STATE_CONNECTED)
 		return SHOUTERR_UNCONNECTED;
 
 	self->error = SHOUTERR_SUCCESS;
 
-	while(remaining) {
-		ret = sock_write_bytes(self->socket, data, remaining);
-		if(ret == (ssize_t)remaining)
-			return len;
-		else if(ret < 0) {
-			if(errno == EINTR)
-				ret = 0;
-			else
-				return self->error = SHOUTERR_SOCKET;
+	/* send immediately if possible (should be the common case) */
+	if (len && ! self->wqueue.len) {
+		if ((ret = try_write(self, data, len)) < 0)
+			return self->error;
+		if (ret < len) {
+			self->error = queue_data(&self->wqueue, data + ret, len - ret);
+			if (self->error != SHOUTERR_SUCCESS)
+				return self->error;
 		}
-		remaining -= ret;
+
+		return len;
 	}
 
-	return len;
+	self->error = queue_data(&self->wqueue, data, len);
+	if (self->error != SHOUTERR_SUCCESS)
+		return self->error;
+
+	ret = send_queue(self);
+	if (ret == SHOUTERR_SUCCESS || (len && ret == SHOUTERR_BUSY))
+		return len;
+
+	return ret;
 }
 
+ssize_t shout_queuelen(shout_t *self)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	return (ssize_t)self->wqueue.len;
+}
+
+
 void shout_sync(shout_t *self)
 {
 	int64_t sleep;
@@ -255,14 +247,15 @@ void shout_sync(shout_t *self)
 	if (self->senttime == 0)
 		return;
 
-	sleep = ((double)self->senttime / 1000) - (timing_get_time() - self->starttime);
-
+	sleep = self->senttime / 1000 - (timing_get_time() - self->starttime);
 	if (sleep > 0)
 		timing_sleep((uint64_t)sleep);
+		
 }
 
 int shout_delay(shout_t *self)
 {
+
 	if (!self)
 		return 0;
 
@@ -270,7 +263,7 @@ int shout_delay(shout_t *self)
 		return 0;
 
 	/* Is this cast to double needed? */
-	return (double)self->senttime / 1000 - (timing_get_time() - self->starttime);
+	return self->senttime / 1000 - (timing_get_time() - self->starttime);
 }
   
 shout_metadata_t *shout_metadata_new(void)
@@ -373,6 +366,8 @@ const char *shout_get_error(shout_t *self)
 		return "Cannot set parameter while connected";
 	case SHOUTERR_UNCONNECTED:
 		return "Not connected";
+        case SHOUTERR_BUSY:
+                return "Socket is busy";
 	case SHOUTERR_UNSUPPORTED:
 		return "This libshout doesn't support the requested option";
 	default:
@@ -380,12 +375,27 @@ const char *shout_get_error(shout_t *self)
 	}
 }
 
+/* Returns:
+ *   SHOUTERR_CONNECTED if the connection is open,
+ *   SHOUTERR_UNCONNECTED if it has not yet been opened,
+ *   or an error from try_connect, including SHOUTERR_BUSY
+ */
 int shout_get_connected(shout_t *self)
 {
-	if (self->connected)
+	int rc;
+
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	if (self->state == SHOUT_STATE_CONNECTED)
 		return SHOUTERR_CONNECTED;
-	else
-		return SHOUTERR_UNCONNECTED;
+	if (self->state != SHOUT_STATE_UNCONNECTED) {
+		if ((rc = try_connect(self)) == SHOUTERR_SUCCESS)
+			return SHOUTERR_CONNECTED;
+		return rc;
+	}
+
+	return SHOUTERR_UNCONNECTED;
 }
 
 int shout_set_host(shout_t *self, const char *host)
@@ -393,7 +403,7 @@ int shout_set_host(shout_t *self, const char *host)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->host)
@@ -418,7 +428,7 @@ int shout_set_port(shout_t *self, unsigned short port)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	self->port = port;
@@ -439,7 +449,7 @@ int shout_set_password(shout_t *self, const char *password)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->password)
@@ -466,7 +476,7 @@ int shout_set_mount(shout_t *self, const char *mount)
 	if (!self || !mount)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 	
 	if (self->mount)
@@ -497,7 +507,7 @@ int shout_set_name(shout_t *self, const char *name)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->name)
@@ -522,7 +532,7 @@ int shout_set_url(shout_t *self, const char *url)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->url)
@@ -547,7 +557,7 @@ int shout_set_genre(shout_t *self, const char *genre)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->genre)
@@ -572,7 +582,7 @@ int shout_set_agent(shout_t *self, const char *agent)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->useragent)
@@ -598,7 +608,7 @@ int shout_set_user(shout_t *self, const char *username)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->user)
@@ -623,7 +633,7 @@ int shout_set_description(shout_t *self, const char *description)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (self->description)
@@ -648,7 +658,7 @@ int shout_set_dumpfile(shout_t *self, const char *dumpfile)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return SHOUTERR_CONNECTED;
 
 	if (self->dumpfile)
@@ -683,7 +693,7 @@ int shout_set_public(shout_t *self, unsigned int public)
 	if (!self || (public != 0 && public != 1))
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	self->public = public;
@@ -704,10 +714,10 @@ int shout_set_format(shout_t *self, unsigned int format)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
-	if (format != SHOUT_FORMAT_VORBIS && format != SHOUT_FORMAT_MP3)
+	if (format != SHOUT_FORMAT_OGG && format != SHOUT_FORMAT_MP3)
 		return self->error = SHOUTERR_UNSUPPORTED;
 
 	self->format = format;
@@ -728,7 +738,7 @@ int shout_set_protocol(shout_t *self, unsigned int protocol)
 	if (!self)
 		return SHOUTERR_INSANE;
 
-	if (self->connected)
+	if (self->state != SHOUT_STATE_UNCONNECTED)
 		return self->error = SHOUTERR_CONNECTED;
 
 	if (protocol != SHOUT_PROTOCOL_HTTP &&
@@ -749,69 +759,386 @@ unsigned int shout_get_protocol(shout_t *self)
 	return self->protocol;
 }
 
+int shout_set_nonblocking(shout_t *self, unsigned int nonblocking)
+{
+	if (!self || (nonblocking != 0 && nonblocking != 1))
+		return SHOUTERR_INSANE;
+
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
+
+	self->nonblocking = nonblocking;
+
+	return SHOUTERR_SUCCESS;
+}
+
+unsigned int shout_get_nonblocking(shout_t *self)
+{
+	if (!self)
+		return 0;
+
+	return self->nonblocking;
+}
+
 /* -- static function definitions -- */
 
-static int send_http_request(shout_t *self, char *username, char *password)
+/* queue data in pages of SHOUT_BUFSIZE bytes */
+static int queue_data(shout_queue_t *queue, const unsigned char *data, size_t len)
+{
+	shout_buf_t *buf;
+	size_t plen;
+
+	if (!len)
+		return SHOUTERR_SUCCESS;
+
+	if (!queue->len) {
+		queue->head = calloc(1, sizeof (shout_buf_t));
+		if (! queue->head)
+			return SHOUTERR_MALLOC;
+	}
+
+	for (buf = queue->head; buf->next; buf = buf->next);
+
+	/* Maybe any added data should be freed if we hit a malloc error?
+	 * Otherwise it'd be impossible to tell where to start requeueing.
+	 * (As if anyone ever tried to recover from a malloc error.) */
+	while (len > 0) {
+		if (buf->len == SHOUT_BUFSIZE) {
+			buf->next = calloc(1, sizeof (shout_buf_t));
+			if (! buf->next)
+				return SHOUTERR_MALLOC;
+			buf->next->prev = buf;
+			buf = buf->next;
+		}
+
+		plen = len > SHOUT_BUFSIZE - buf->len ? SHOUT_BUFSIZE - buf->len : len;
+		memcpy (buf->data + buf->len, data, plen);
+		buf->len += plen;
+		data += plen;
+		len -= plen;
+		queue->len += plen;
+	}
+
+	return SHOUTERR_SUCCESS;
+}
+
+static inline int queue_str(shout_t *self, const char *str)
+{
+	return queue_data(&self->wqueue, str, strlen(str));
+}
+
+/* this should be shared with sock_write. Create libicecommon. */
+static int queue_printf(shout_t *self, const char *fmt, ...)
+{
+	char buffer[1024];
+	char *buf;
+	va_list ap, ap_retry;
+	int len;
+
+	buf = buffer;
+
+	va_start(ap, fmt);
+	va_copy(ap_retry, ap);
+
+	len = vsnprintf(buf, sizeof(buffer), fmt, ap);
+
+	self->error = SHOUTERR_SUCCESS;
+	if (len > 0) {
+		if ((size_t)len < sizeof(buffer))
+			queue_data(&self->wqueue, buf, len);
+		else {
+			buf = malloc(++len);
+			if (buf) {
+				len = vsnprintf(buf, len, fmt, ap_retry);
+				queue_data(&self->wqueue, buf, len);
+				free(buf);
+			} else
+				self->error = SHOUTERR_MALLOC;
+		}
+	}
+
+	va_end(ap_retry);
+	va_end(ap);
+
+	return self->error;
+}
+
+static inline void queue_free(shout_queue_t *queue)
+{
+	shout_buf_t *prev;
+
+	while (queue->head) {
+		prev = queue->head;
+		queue->head = queue->head->next;
+		free(prev);
+	}
+	queue->len = 0;
+}
+
+static int get_response(shout_t *self)
+{
+	char buf[1024];
+	int rc, blen;
+	char *pc;
+	shout_buf_t *queue;
+	int newlines = 0;
+
+	rc = sock_read_bytes(self->socket, buf, sizeof(buf));
+
+	if (rc < 0 && sock_recoverable(sock_error()))
+		return SHOUTERR_BUSY;
+	if (rc <= 0)
+		return SHOUTERR_SOCKET;
+
+	if ((rc = queue_data(&self->rqueue, buf, rc)))
+		return rc;
+
+	/* work from the back looking for \r?\n\r?\n. Anything else means more
+	 * is coming. */
+	for (queue = self->rqueue.head; queue->next; queue = queue->next);
+	pc = queue->data + queue->len - 1;
+	blen = queue->len;
+	while (blen) {
+		if (*pc == '\n')
+			newlines++;
+		else if (*pc != '\r')
+			break;
+
+		if (newlines == 2)
+			return SHOUTERR_SUCCESS;
+
+		blen--;
+		pc--;
+
+		if (!blen && queue->prev) {
+			queue = queue->prev;
+			pc = queue->data + queue->len - 1;
+			blen = queue->len;
+		}
+	}
+
+	return SHOUTERR_BUSY;
+}
+
+static int try_connect (shout_t *self)
+{
+	int rc;
+	int port;
+
+	/* the breaks between cases are omitted intentionally */
+	switch (self->state) {
+	case SHOUT_STATE_UNCONNECTED:
+		port = self->port;
+		if (shout_get_protocol(self) == SHOUT_PROTOCOL_ICY)
+			port++;
+
+		if (shout_get_nonblocking(self)) {
+			if ((self->socket = sock_connect_non_blocking(self->host, port)) < 0)
+				return self->error = SHOUTERR_NOCONNECT;
+			self->state = SHOUT_STATE_CONNECT_PENDING;
+		} else {
+			if ((self->socket = sock_connect(self->host, port)) < 0)
+				return self->error = SHOUTERR_NOCONNECT;
+			if ((rc = create_request(self)) != SHOUTERR_SUCCESS)
+				return rc;
+			self->state = SHOUT_STATE_REQ_PENDING;
+		}
+
+	case SHOUT_STATE_CONNECT_PENDING:
+		if (shout_get_nonblocking(self)) {
+			if ((rc = sock_connected(self->socket, 0)) < 1) {
+				if (!rc)
+					return SHOUTERR_BUSY;
+				else
+					return SHOUTERR_SOCKET;
+			}
+			if ((rc = create_request(self)) != SHOUTERR_SUCCESS)
+				return rc;
+		}
+		self->state = SHOUT_STATE_REQ_PENDING;
+
+	case SHOUT_STATE_REQ_PENDING:
+		do
+			rc = send_queue(self);
+		while (!shout_get_nonblocking(self) && rc == SHOUTERR_BUSY);
+		if (rc != SHOUTERR_SUCCESS)
+			return rc;
+		self->state = SHOUT_STATE_RESP_PENDING;
+
+	case SHOUT_STATE_RESP_PENDING:
+		do
+			rc = get_response(self);
+		while (!shout_get_nonblocking(self) && rc == SHOUTERR_BUSY);
+		if (rc != SHOUTERR_SUCCESS)
+			return rc;
+
+		if ((rc = parse_response(self)) != SHOUTERR_SUCCESS)
+			return rc;
+
+		if (self->format == SHOUT_FORMAT_OGG) {
+			if ((self->error = shout_open_ogg(self)) != SHOUTERR_SUCCESS) {
+				self->state = SHOUT_STATE_UNCONNECTED;
+				sock_close(self->socket);
+				return self->error;
+			}
+		} else if (self->format == SHOUT_FORMAT_MP3) {
+			if ((self->error = shout_open_mp3(self)) != SHOUTERR_SUCCESS) {
+				self->state = SHOUT_STATE_UNCONNECTED;
+				sock_close(self->socket);
+				return self->error;
+			}
+		} else {
+			self->state = SHOUT_STATE_UNCONNECTED;
+			sock_close(self->socket);
+			return self->error = SHOUTERR_INSANE;
+		}
+
+	case SHOUT_STATE_CONNECTED:
+		self->state = SHOUT_STATE_CONNECTED;
+	}
+	
+	return SHOUTERR_SUCCESS;
+}
+
+static int try_write (shout_t *self, const void *data, size_t len)
+{
+    int ret;
+    size_t pos = 0;
+
+    /* loop until whole buffer is written (unless it would block) */
+    do {
+        ret = sock_write_bytes (self->socket, data + pos, len - pos);
+        if (ret > 0)
+            pos += ret;
+    } while (pos < len && ret >= 0);
+
+    if (ret < 0)
+    {
+        if (sock_recoverable (sock_error()))
+        {
+            self->error = SHOUTERR_BUSY;
+            return 0;
+        }
+        self->error = SHOUTERR_SOCKET;
+        return ret;
+    }
+
+    return pos;
+}
+
+/* collect nodes of a queue into a single buffer */
+static int collect_queue(shout_buf_t *queue, char **buf)
+{
+	shout_buf_t *node;
+	int pos = 0;
+	int len = 0;
+
+	for (node = queue; node; node = node->next)
+		len += node->len;
+
+	if (!(*buf = malloc(len)))
+		return SHOUTERR_MALLOC;
+
+	for (node = queue; node; node = node->next) {
+		memcpy(*buf + pos, node->data, node->len);
+		pos += node->len;
+	}
+
+	return len;
+}
+
+static int send_queue(shout_t *self)
+{
+	shout_buf_t *buf;
+	int ret;
+
+	if (!self->wqueue.len)
+		return SHOUTERR_SUCCESS;
+
+	buf = self->wqueue.head;
+	while (buf) {
+		ret = try_write (self, buf->data + buf->pos, buf->len - buf->pos);
+		if (ret < 0)
+			return self->error;
+
+		buf->pos += ret;
+		self->wqueue.len -= ret;
+		if (buf->pos == buf->len) {
+			self->wqueue.head = buf->next;
+			free(buf);
+			buf = self->wqueue.head;
+			if (buf)
+				buf->prev = NULL;
+		} else /* incomplete write */
+			return SHOUTERR_BUSY;
+	}
+
+	return self->error = SHOUTERR_SUCCESS;
+}
+
+static int create_request(shout_t *self)
+{
+	if (self->protocol == SHOUT_PROTOCOL_HTTP)
+		return create_http_request(self);
+	else if (self->protocol == SHOUT_PROTOCOL_XAUDIOCAST)
+		return create_xaudiocast_request(self);
+	else if (self->protocol == SHOUT_PROTOCOL_ICY)
+		return create_icy_request(self);
+
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+
+static int create_http_request(shout_t *self)
 {
 	char *auth;
 	char *ai;
+	int ret = SHOUTERR_MALLOC;
 
-	if (!sock_write(self->socket, "SOURCE %s HTTP/1.0\r\n", self->mount))
-		return SHOUTERR_SOCKET;
-
-	if (self->password && (auth = http_basic_authorization(self))) {
-		if (!sock_write(self->socket, auth)) {
+	/* this is lazy code that relies on the only error from queue_* being
+	 * SHOUTERR_MALLOC */
+	do {
+		if (queue_printf(self, "SOURCE %s HTTP/1.0\r\n", self->mount))
+			break;
+		if (self->password) {
+			if (! (auth = http_basic_authorization(self)))
+				break;
+			if (queue_str(self, auth)) {
+				free(auth);
+				break;
+			}
 			free(auth);
-			return SHOUTERR_SOCKET;
 		}
-		free(auth);
-	}
+		if (self->useragent && queue_printf(self, "User-Agent: %s\r\n", self->useragent))
+			break;
+		if (self->format == SHOUT_FORMAT_OGG && queue_printf(self, "Content-Type: application/ogg\r\n"))
+			break;
+		if (self->format == SHOUT_FORMAT_MP3 && queue_printf(self, "Content-Type: audio/mpeg\r\n"))
+			break;
+		if (queue_printf(self, "ice-name: %s\r\n", self->name ? self->name : "no name"))
+			break;
+		if (queue_printf(self, "ice-public: %d\r\n", self->public))
+			break;
 
-	if (!sock_write(self->socket, "ice-name: %s\r\n", self->name != NULL ? self->name : "no name"))
-		return SHOUTERR_SOCKET;
-	if (self->url) {
-		if (!sock_write(self->socket, "ice-url: %s\r\n", self->url))
-			return SHOUTERR_SOCKET;
-	}
-	if (self->genre) {
-		if (!sock_write(self->socket, "ice-genre: %s\r\n", self->genre))
-			return SHOUTERR_SOCKET;
-	}
-#if 0
-	ai = shout_get_audio_info(self, SHOUT_AI_BITRATE);
-
-	if (bitrate && !sock_write(self->socket, "ice-bitrate: %s\r\n", bitrate))
-		return SHOUTERR_SOCKET;
-#else
-	if ((ai = _shout_util_dict_urlencode(self->audio_info, ';'))) {
-		if (!sock_write(self->socket, "ice-audio-info: %s\r\n", ai)) {
+		if (self->url && queue_printf(self, "ice-url: %s\r\n", self->url))
+			break;
+		if (self->genre && queue_printf(self, "ice-genre: %s\r\n", self->genre))
+			break;
+		if ((ai = _shout_util_dict_urlencode(self->audio_info, ';'))) {
+			if (queue_printf(self, "ice-audio-info: %s\r\n", ai)) {
+				free(ai);
+				break;
+			}
 			free(ai);
-			return SHOUTERR_SOCKET;
 		}
-	}
-#endif
-	if (!sock_write(self->socket, "ice-public: %d\r\n", self->public))
-		return SHOUTERR_SOCKET;
-	if (self->description) {
-		if (!sock_write(self->socket, "ice-description: %s\r\n", self->description))
-			return SHOUTERR_SOCKET;
-	}
-	if (self->useragent) {
-		if (!sock_write(self->socket, "User-Agent: %s\r\n", self->useragent))
-			return SHOUTERR_SOCKET;
-	}
-	if (self->format == SHOUT_FORMAT_VORBIS) {
-		if (!sock_write(self->socket, "Content-Type: application/ogg\r\n"))
-			return SHOUTERR_SOCKET;
-	} else if (self->format == SHOUT_FORMAT_MP3) {
-		if (!sock_write(self->socket, "Content-Type: audio/mpeg\r\n"))
-			return SHOUTERR_SOCKET;
-	}
+		if (self->description && queue_printf(self, "ice-description: %s\r\n", self->description))
+			break;
+		if (queue_str(self, "\r\n"))
+			break;
+		
+		ret = SHOUTERR_SUCCESS;
+	} while (0);
 
-	if (!sock_write(self->socket, "\r\n"))
-		return SHOUTERR_SOCKET;
-
-	return SHOUTERR_SUCCESS;
+	return ret;
 }
 
 static char *http_basic_authorization(shout_t *self)
@@ -840,162 +1167,132 @@ static char *http_basic_authorization(shout_t *self)
 	return in;
 }
 
-static int login_http_basic(shout_t *self)
+static int parse_response(shout_t *self)
+{
+	if (self->protocol == SHOUT_PROTOCOL_HTTP)
+		return parse_http_response(self);
+	else if (self->protocol == SHOUT_PROTOCOL_XAUDIOCAST ||
+		 self->protocol == SHOUT_PROTOCOL_ICY)
+		return parse_xaudiocast_response(self);
+
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+
+static int parse_http_response(shout_t *self)
 {
-	char header[4096];
 	http_parser_t *parser;
+	char *header;
+	int hlen = 0;
 	int code;
 	char *retcode;
 #if 0
 	char *realm;
 #endif
-    
-	self->error = SHOUTERR_SOCKET;
 
-   	self->socket = sock_connect(self->host, self->port);
-	if (self->socket < 0) {
-		return self->error = SHOUTERR_NOCONNECT;
-	}
-
-#if 0
-	if(send_http_request(self, NULL, NULL) != 0) {
-#else
-	/* assume we'll have to authenticate, saves round trips on basic */
-	if(send_http_request(self, self->user, self->password) != 0) {
-#endif
-		return self->error = SHOUTERR_SOCKET;
-	}
-
-	if (_shout_util_read_header(self->socket, header, 4096) == 0) {
-		/* either we didn't get a complete header, or we timed out */
-		return self->error = SHOUTERR_SOCKET;
-	}
+	/* all this copying! */
+	hlen = collect_queue(self->rqueue.head, &header);
+	if (hlen <= 0)
+		return SHOUTERR_MALLOC;
+	queue_free(&self->rqueue);
 
 	parser = httpp_create_parser();
 	httpp_initialize(parser, NULL);
-	if (httpp_parse_response(parser, header, strlen(header), self->mount)) {
+	if (httpp_parse_response(parser, header, hlen, self->mount)) {
+		free (header);
 		retcode = httpp_getvar(parser, HTTPP_VAR_ERROR_CODE);
 		code = atoi(retcode);
 		if(code >= 200 && code < 300) {
 			httpp_destroy(parser);
 			return SHOUTERR_SUCCESS;
 		}
-#if 0
-		else if(code == 401) {
-			/* Don't really use this right now other than to check that it's
-			* present.
-			*/
-			realm = httpp_getvar(parser, "www-authenticate");
-			if(realm) {
-				httpp_destroy(parser);
-				sock_close(self->socket);
-
-				self->socket = sock_connect(self->host, self->port);
-				if (self->socket <= 0)
-					return self->error = SHOUTERR_NOCONNECT;
-
-				if(send_http_request(self, self->user, self->password) != 0) {
-					sock_close(self->socket);
-					return self->error = SHOUTERR_SOCKET;
-				}
-
-				if (_shout_util_read_header(self->socket, header, 4096) == 0) {
-					/* either we didn't get a complete header, or we timed out */
-					sock_close(self->socket);
-					return self->error = SHOUTERR_SOCKET;
-				}
-				parser = httpp_create_parser();
-				httpp_initialize(parser, NULL);
-				if (httpp_parse_response(parser, header, strlen(header), self->mount)) {
-					retcode = httpp_getvar(parser, HTTPP_VAR_ERROR_CODE);
-					code = atoi(retcode);
-					if(code >= 200 && code < 300) {
-						httpp_destroy(parser);
-						return SHOUTERR_SUCCESS;
-					}
-				}
-			}
-		}
-#endif
 	}
 
+	free(header);
 	httpp_destroy(parser);
 	return self->error = SHOUTERR_NOLOGIN;
 }
 
-static int login_xaudiocast(shout_t *self)
+static int create_xaudiocast_request(shout_t *self)
 {
-	char response[4096];
 	const char *bitrate;
+	int ret;
 
 	bitrate = shout_get_audio_info(self, SHOUT_AI_BITRATE);
 	if (!bitrate)
 		bitrate = "0";
 
-	if (!sock_write(self->socket, "SOURCE %s %s\n", self->password, self->mount))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-name: %s\n", self->name != NULL ? self->name : "unnamed"))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-url: %s\n", self->url != NULL ? self->url : "http://www.icecast.org/"))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-genre: %s\n", self->genre != NULL ? self->genre : "icecast"))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-bitrate: %s\n", bitrate))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-public: %i\n", self->public))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "x-audiocast-description: %s\n", self->description != NULL ? self->description : "Broadcasting with the icecast streaming media server!"))
-		return SHOUTERR_SOCKET;
-	if (self->dumpfile && !sock_write(self->socket, "x-audiocast-dumpfile: %s\n", self->dumpfile))
-		return SHOUTERR_SOCKET;
+	ret = SHOUTERR_MALLOC;
+	do {
+		if (queue_printf(self, "SOURCE %s %s\n", self->password, self->mount))
+			break;
+		if (queue_printf(self, "x-audiocast-name: %s\n", self->name ? self->name : "unnamed"))
+			break;
+		if (queue_printf(self, "x-audiocast-url: %s\n", self->url ? self->url : "http://www.icecast.org/"))
+			break;
+		if (queue_printf(self, "x-audiocast-genre: %s\n", self->genre ? self->genre : "icecast"))
+			break;
+		if (queue_printf(self, "x-audiocast-bitrate: %s\n", bitrate))
+			break;
+		if (queue_printf(self, "x-audiocast-public: %i\n", self->public))
+			break;
+		if (queue_printf(self, "x-audiocast-description: %s\n", self->description ? self->description : "Broadcasting with the icecast streaming media server!"))
+			break;
+		if (self->dumpfile && queue_printf(self, "x-audiocast-dumpfile: %s\n", self->dumpfile))
+			break;
+		if (queue_str(self, "\n"))
+			break;
 
-	if (!sock_write(self->socket, "\n"))
-		return SHOUTERR_SOCKET;
+		ret = SHOUTERR_SUCCESS;
+	} while (0);
+		
+	return ret;
+}
 
-	if (!sock_read_line(self->socket, response, sizeof(response)))
-		return SHOUTERR_SOCKET;
+static int parse_xaudiocast_response(shout_t *self)
+{
+	char *response;
 
-	if (!strstr(response, "OK"))
+	if (collect_queue(self->rqueue.head, &response) <= 0)
+		return SHOUTERR_MALLOC;
+	queue_free(&self->rqueue);
+
+	if (!strstr(response, "OK")) {
+		free(response);
 		return SHOUTERR_NOLOGIN;
+	}
+	free(response);
 
 	return SHOUTERR_SUCCESS;
 }
 
-int login_icy(shout_t *self)
+static int create_icy_request(shout_t *self)
 {
-	char response[4096];
 	const char *bitrate;
+	int ret;
 
 	bitrate = shout_get_audio_info(self, SHOUT_AI_BITRATE);
 	if (!bitrate)
 		bitrate = "0";
 
-	if (!sock_write(self->socket, "%s\n", self->password))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "icy-name:%s\n", self->name != NULL ? self->name : "unnamed"))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "icy-url:%s\n", self->url != NULL ? self->url : "http://www.icecast.org/"))
-		return SHOUTERR_SOCKET;
+	ret = SHOUTERR_MALLOC;
+	do {
+		if (queue_printf(self, "%s\n", self->password))
+			break;
+		if (queue_printf(self, "icy-name:%s\n", self->name ? self->name : "unnamed"))
+			break;
+		if (queue_printf(self, "icy-url:%s\n", self->url ? self->url : "http://www.icecast.org/"))
+			break;
+		if (queue_str(self, "icy-irc:\nicy-aim:icy-icq:\n"))
+			break;
+		if (queue_printf(self, "icy-pub:%i\n", self->public))
+			break;
+		if (queue_printf(self, "icy-genre:%s\n", self->genre ? self->genre : "icecast"))
+			break;
+		if (queue_printf(self, "icy-br:%s\n\n", bitrate))
+			break;
 
-	/* Fields we don't use */
-	if (!sock_write(self->socket, "icy-irc:\nicy-aim:\nicy-icq:\n"))
-		return SHOUTERR_SOCKET;
+		ret = SHOUTERR_SUCCESS;
+	} while (0);
 
-	if (!sock_write(self->socket, "icy-pub:%i\n", self->public))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "icy-genre:%s\n", self->genre != NULL ? self->genre : "icecast"))
-		return SHOUTERR_SOCKET;
-	if (!sock_write(self->socket, "icy-br:%s\n", bitrate))
-		return SHOUTERR_SOCKET;
-
-	if (!sock_write(self->socket, "\n"))
-		return SHOUTERR_SOCKET;
-
-	if (!sock_read_line(self->socket, response, sizeof(response)))
-		return SHOUTERR_SOCKET;
-
-	if (!strstr(response, "OK"))
-		return SHOUTERR_NOLOGIN;
-
-	return SHOUTERR_SUCCESS;
+	return ret;
 }
